#include "BPQConfigGen.h"
#ifdef WIN32
#include <Windows.h>
#else

#define TRUE 1
#define FALSE 0

#define strtok_s strtok_r

int memicmp(unsigned char *a, unsigned char *b, int n)
{
	if (n)
	{
		while (n && toupper(*a) == toupper(*b))
			n--, a++, b++;

		if (n)
			return toupper(*a) - toupper(*b);
	}
	return 0;
}
int stricmp(const char * pStr1, const char *pStr2)
{
	unsigned char c1, c2;
	int  v;

	if (pStr1 == NULL)
		return 1;
	
	do {
		c1 = *pStr1++;
		c2 = *pStr2++;
		/* The casts are necessary when pStr1 is shorter & char is signed */
		v = tolower(c1) - tolower(c2);
	} while ((v == 0) && (c1 != '\0') && (c2 != '\0'));

	return v;
}
#endif

char * strlop(char * buf, char delim)
{
	// Terminate buf at delim, and return rest of string

	char * ptr = strchr(buf, delim);

	if (ptr == NULL) return NULL;

	*(ptr)++ = 0;

	return ptr;
}

void portSetVisible(int i, int v);
void applSetVisible(int i, int v);
void refreshPorts();
void refreshAppls();

bool HadConfig = false;		// Set if we've read a config 

int BBSAppl = 0;
int ChatAppl = 0;
int RMSAppl = 0;
int TelnetPort = 0;

QAction * helpAction;

QString NodeCall;
QString NodeAlias;
QString LOC;

QString CMSCall;
QString CMSPass;
QString User;
QString Pass;
char LoppedCall[80];
int NodeSSID = 0;

QPoint mainpos;

QLineEdit * sysopUser;
QLineEdit * sysopPass;

int RMS;


char * NODECALL = NULL;
char * NODEALIAS = NULL;

int nextPortnum = 0;
int activePortnum = 0;	// May be different from above if PORTNUM used

char ValfromReg[250] = "";

QLabel * gPortNum[32];
QLabel * gPortName[32];
QLabel * gPortType[32];
QPushButton * gPortEdit[32];

QLineEdit * gapplNum[32];
QLineEdit * gapplCmd[32];
QLineEdit * gapplCall[32];
QLineEdit * gapplAlias[32];
QLineEdit * gapplCmdAlias[32];
QLineEdit * gapplQual[32];
QPushButton * gapplEdit[32];

int PortType[32] = { 0 };

char ID[32][33] = { "" };
char Param1[32][32] = { "" };
int Param2[32] =  {0};
char Param3[32][32] = { "" };

char * portConfig[32] = { nullptr };	// Malloc'ed Config

#define H_AEA 0
#define H_ARDOP 1
#define H_AXIP 2
#define H_ETHER 3
#define H_FLDIGI 4
#define H_HAL 5
#define H_KAM 6
#define H_LOOP 7
#define H_KISS 8 
#define H_KISSTCP 9
#define H_KISSUDP 10
#define H_MPSK 11
#define H_SCS 12
#define H_TRK 13
#define H_TRKM 14
#define H_SERIAL 15
#define H_TELNET 16
#define H_UZ7HO 17
#define H_V4 18
//#define H_UIARQ 12
#define H_VARA 19
#define H_VKISS 20
#define H_WINMOR 21

#define NUMBEROFTYPES 22

char Types[NUMBEROFTYPES][14] = {
	"AEA Pactor", "ARDOP", "AXIP", "BPQETHER",
	"FLDIGI", "HAL", "KAM Pactor", "LOOPBACK",
	"KISS Serial", "KISS TCP", "KISS UDP", "MULTIPSK", 
	"SCS Pactor", "SCS Tracker","SCS TRKMULTI","Serial",
	"TELNET", "UZ7HO", "V4", "VARA",
	"VKISS", "WINMOR"};

int defaultPort[NUMBEROFTYPES] = {
	9600, 8515, 0, 0, 0, 9600, 9600, 0,
	9600, 8100, 8100, 0, 38400, 38400, 38400, 9600,
	0, 8000, 8510, 8300, 0, 8500};

struct APPL
{
	int Num;
	char Command[12];
	char CommandAlias[48];
	char ApplCall[10];
	char ApplAlias[10];
	char L2Alias[10];
	int ApplQual;
};



struct APPL Appls[33] = { 0, "", "", "", "", "", 0};

int nextApplNum = 0;

char mainConfig[32768] = "";

// Telnet Server User Record

struct UserRec
{
	char * Callsign;
	char * UserName;
	char * Password;
	char * Appl;				// Autoconnect APPL
	int Secure;				    // SYSOP flag

	QLineEdit * gCallsign;
	QLineEdit * gUserName;
	QLineEdit * gPassword;
	QLineEdit * gAppl;
	QCheckBox * gSYSOP;

};

struct UserRec Users[1024] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

int NextUser = 0;

QScrollArea *userscrollArea;
QWidget *userScroll;
QLabel * UserLabel;

QWidget * placeHolder;


BPQConfigGen::BPQConfigGen(QWidget *parent)
	: QMainWindow(parent)
{
	int i;
	char Label[10];
	ui.setupUi(this);

	// get local copies of some controls

	sysopUser = ui.sysopUser;
	sysopPass = ui.sysopPass;

	placeHolder = new QWidget();

	// Set up port list

	for (i = 0; i < 32; i++)
	{
		sprintf(Label, "%d", i + 1);
		gPortNum[i] = new QLabel(Label, ui.portScroll);
		gPortNum[i]->setGeometry(QRect(5, 22 * i + 5, 20, 18));

		gPortType[i] = new QLabel(ui.portScroll);
		gPortType[i]->setGeometry(QRect(25, 22 * i + 5, 100, 18));

		gPortName[i] = new QLabel(ui.portScroll);
		gPortName[i]->setGeometry(QRect(130, 22 * i + 5, 250, 18));

		gPortEdit[i] = new QPushButton("Edit", ui.portScroll);
		gPortEdit[i]->setGeometry(QRect(385, 22 * i + 5, 60, 18));
		connect(gPortEdit[i], SIGNAL(clicked()), this, SLOT(clickedSlot()));

		portSetVisible(i, 0);
	}

	for (i = 0; i < 32; i++)
	{
		sprintf(Label, "%d", i + 1); 
		gapplNum[i] = new QLineEdit(Label, ui.applScroll);
		gapplNum[i]->setGeometry(QRect(5, 22 * i + 5, 20, 18));
		gapplCmd[i] = new QLineEdit(ui.applScroll);
		gapplCmd[i]->setGeometry(QRect(25, 22 * i + 5, 100, 18));
		gapplCmdAlias[i] = new QLineEdit(ui.applScroll);
		gapplCmdAlias[i]->setGeometry(QRect(130, 22 * i + 5, 100, 18));
		gapplCall[i] = new QLineEdit(ui.applScroll);
		gapplCall[i]->setGeometry(QRect(235, 22 * i + 5, 100, 18));
		gapplAlias[i] = new QLineEdit(ui.applScroll);
		gapplAlias[i]->setGeometry(QRect(340, 22 * i + 5, 100, 18));
		gapplQual[i] = new QLineEdit(ui.applScroll);
		gapplQual[i]->setGeometry(QRect(445, 22 * i + 5, 30, 18));
		gapplEdit[i] = new QPushButton("Save", ui.applScroll);
		gapplEdit[i]->setGeometry(QRect(480, 22 * i + 5, 60, 18));
		connect(gapplEdit[i], SIGNAL(clicked()), this, SLOT(clickedSlot()));

		applSetVisible(i, 0);
	}

	// Get default BPQ Directory

#ifdef Q_OS_WIN32

	HKEY REGTREE = HKEY_CURRENT_USER;

	HKEY hKey = 0;
	HKEY hKeyIn = 0;
	HKEY hKeyOut = 0;
	int disp;
	int retCode;
	DWORD Type, Vallen = MAX_PATH;
	char msg[512];
	char DLLName[256] = "Not Known";
	char LogDir[256];

	retCode = RegOpenKeyExA(REGTREE,
		"SOFTWARE\\G8BPQ\\BPQ32",
		0,
		KEY_QUERY_VALUE,
		&hKey);

	if (retCode == ERROR_SUCCESS)
	{
		// Try "BPQ Directory"

		Vallen = MAX_PATH;
		retCode = RegQueryValueExA(hKey, "BPQ Directory", 0,
			&Type, (UCHAR *)&ValfromReg, &Vallen);

		if (retCode == ERROR_SUCCESS)
		{
			if (strlen(ValfromReg) == 2 && ValfromReg[0] == '"' && ValfromReg[1] == '"')
				ValfromReg[0] = 0;
		}

		if (ValfromReg[0] == 0)
		{
			// BPQ Directory absent or = "" - try "Config File Location"

			Vallen = MAX_PATH;

			retCode = RegQueryValueExA(hKey, "Config File Location", 0,
				&Type, (UCHAR *)&ValfromReg, &Vallen);

			if (retCode == ERROR_SUCCESS)
			{
				if (strlen(ValfromReg) == 2 && ValfromReg[0] == '"' && ValfromReg[1] == '"')
					ValfromReg[0] = 0;
			}
		}
	}

	if (ValfromReg[0] == 0)
		GetCurrentDirectoryA(MAX_PATH, ValfromReg);

	ui.Directory->setText(ValfromReg);

#else
	char * dir = getenv("PWD");
	ui.Directory->setText(dir);
#endif

	connect(ui.ReadConfig, SIGNAL(clicked()), this, SLOT(clickedSlot()));
	connect(ui.Update, SIGNAL(clicked()), this, SLOT(clickedSlot()));
	connect(ui.ADDPORT, SIGNAL(clicked()), this, SLOT(clickedSlot()));
	connect(ui.ADDAPP, SIGNAL(clicked()), this, SLOT(clickedSlot()));
	connect(ui.EDITOTHER, SIGNAL(clicked()), this, SLOT(clickedSlot()));

	connect(ui.WEBMGMT, SIGNAL(clicked()), this, SLOT(clickedSlot()));
	connect(ui.BBS, SIGNAL(clicked()), this, SLOT(clickedSlot()));
	connect(ui.CHAT, SIGNAL(clicked()), this, SLOT(clickedSlot()));
	connect(ui.RMS, SIGNAL(clicked()), this, SLOT(clickedSlot()));

	helpAction = ui.menuBar->addAction("Help");
	helpAction->setObjectName(QString::fromUtf8("HELP"));
	connect(helpAction, SIGNAL(triggered()), this, SLOT(clickedSlot()));


	/*
// Create Default Telnet Port for Management

	PortType[0] = H_TELNET;
	strcpy(ID[0], "Telnet Port");

	portConfig[0] = (char *)malloc(1000);
	portConfig[0][0] = 0;

	refreshPorts();
*/
}

void createTelnetConfig(int p)
{
	PortType[p] = H_TELNET;
	strcpy(ID[p], "Telnet Server");

	char * ptr = portConfig[p] = (char *)malloc(1000);

	TelnetPort = ++p;

	ptr += sprintf(ptr, "PORT\r\n");
	ptr += sprintf(ptr, "PORTNUM=%d\r\n", p);
	ptr += sprintf(ptr, " ID=Telnet Server\r\n");
	ptr += sprintf(ptr, " DRIVER=TELNET\r\n");
	ptr += sprintf(ptr, " CONFIG\r\n");
	ptr += sprintf(ptr, "  LOGGING=1\r\n");
	ptr += sprintf(ptr, "  TCPPORT=8010\r\n");
	ptr += sprintf(ptr, "  FBBPORT=8011\r\n");
	ptr += sprintf(ptr, "  HTTPPORT=8080\r\n");

		if (RMS || CMSCall.size() || CMSPass.size())
	{
		ptr += sprintf(ptr, "  CMS=1\r\n");
		if (CMSCall.size())
			ptr += sprintf(ptr, "  CMSCALL=%s\r\n", CMSCall.toLocal8Bit().constData());
		if (CMSPass.size())
			ptr += sprintf(ptr, "  CMSPASS=%s\r\n", CMSPass.toLocal8Bit().constData());
	}
	ptr += sprintf(ptr, "ENDPORT\r\n");
	ptr += sprintf(ptr, "\r\n");
}

void BPQConfigGen::clickedSlot()
{
	QPushButton * Button = static_cast<QPushButton*>(QObject::sender());
	char Name[32];

	mainpos = this->pos();				// for popup windows

	strcpy(Name, sender()->objectName().toUtf8());

	if (strcmp(Name, "HELP") == 0)
	{
		QMessageBox::about(this, tr("Help"),
			"BPQConfigGen is a program to create or edit a BPQ Node configuration file(bpq32.cfg).A minimal configuration can be created by entering a Node Callsign and adding at least one port.All LinBPQ users and most Windows users will want a Telnet port for "
			"management.This can be added by setting a username and password and enabling the Web Mgmt Interface.If you want them you can then enable the BBS, CHAT Server or Winlink Gateway by checking the appropriate box.You may want to customise the Application "
			"definitions by changing the callsign or setting a NETROM Alias and Quality.Other applications can be added using the \"Add Application\" button and filling in the details."
			"<br><br>"
			"BPQConfigGen only allows you to create or edit the basic Node identification info and your ports and application definitions.Other BPQ features, such as APRS, can be added using the \"Edit Main Config\" button and entering the necessary config statements."
			"<br><br>"
			"Once you have a config file you can edit it by using the \"Read Existing Config\" button.This can also read configs not created using BPQConfigGen.If that is the case you will get a warning message that the config may not be read correctly, but normally it will be fine."
			"<br><br>"
			"If you read an existing config file comments will normally be kept, but the order of configuration sections may be changed.The new file will have the Port and Application definitions at the end."
			"<br><br>"
			"Before a file is updated a copy of the original config is saved with a date / time stamp appended.There is no housekeeping of these - you will need to delete any you no longer need.");

		return;
	}

	if (strcmp(Name, "ReadConfig") == 0)
	{
		ReadConfigFile();
		return;
	}

	GetParams();

	// Make sure NodeCall is set before we do anything

	if (NODECALL == 0 || NODECALL[0] == 0)
	{
		NodeCall = ui.NODECALL->text().toUpper();
		NODECALL = strdup(NodeCall.toLocal8Bit().constData());

		strcpy(LoppedCall, NodeCall.toLocal8Bit().constData());

		char * ptr = strchr(LoppedCall, '-');
		if (ptr)
		{
			*ptr++ = 0;
			NodeSSID = atoi(ptr);
		}

		if (NODECALL == 0 || NODECALL[0] == 0)
		{
			QMessageBox msgBox;

			msgBox.setWindowTitle("Error");
			msgBox.setText("Please Set NODECALL before configuring");
			msgBox.move(mainpos.x() + 200, mainpos.y() + 300);
			msgBox.exec();

			Button->setChecked(false);

			return;
		}

	}

	if (strcmp(Name, "ADDPORT") == 0)
	{
		// This runs the Add Port dialog
		
		GetParams();			// Nodecall, etc

		PortDialog dialog(0);
		dialog.exec();

 		return;
	}

	if (strcmp(Name, "ADDAPP") == 0)
	{
		applSetVisible(nextApplNum++, 1);
		return;
	}

	if (strcmp(Name, "EDITOTHER") == 0)
	{
		editOtherDialog dialog(0);
		dialog.exec();

		return;
	}

	if (strcmp(Name, "BBS") == 0)
	{
		// Add BBS

		APPL * pAppl = &Appls[nextApplNum++];

		BBSAppl = pAppl->Num = 1;
		strcpy(pAppl->Command, "BBS");

		strcpy(pAppl->ApplCall, LoppedCall);

		if (strchr(NODECALL, '-') == 0)
			strcat(pAppl->ApplCall, "-1");

		//		strcpy(pAppl->ApplAlias, ApplA);

		pAppl->ApplQual = 0;

		Button->setEnabled(false);

		refreshAppls();

		return;
	}
	if (strcmp(Name, "CHAT") == 0)
	{
		APPL * pAppl = &Appls[nextApplNum++];

		BBSAppl = pAppl->Num = 2;
		strcpy(pAppl->Command, "CHAT");

		strcpy(pAppl->ApplCall, LoppedCall);

		if (NodeSSID == 4)
			strcat(pAppl->ApplCall, "-2");
		else
			strcat(pAppl->ApplCall, "-4");

		//		strcpy(pAppl->ApplAlias, ApplA);

		pAppl->ApplQual = 0;

		Button->setEnabled(false);

		refreshAppls();

		return;
	}
	if (strcmp(Name, "RMS") == 0)
	{
		APPL * pAppl = &Appls[nextApplNum++];

		if (TelnetPort == 0)
		{
			createTelnetConfig(nextPortnum++);
			refreshPorts();
		}

		BBSAppl = pAppl->Num = 3;
		strcpy(pAppl->Command, "RMS");

		strcpy(pAppl->ApplCall, LoppedCall);

		if (NodeSSID == 10)
			strcat(pAppl->ApplCall, "-11");
		else
			strcat(pAppl->ApplCall, "-10");

		sprintf(pAppl->CommandAlias, "C %d CMS", TelnetPort);
		//		strcpy(pAppl->ApplAlias, ApplA);

		pAppl->ApplQual = 0;

		Button->setEnabled(false);

		refreshAppls();

		return;
	}

	if (strcmp(Name, "WEBMGMT") == 0)
	{
		if (User.length() == 0 || Pass.length() == 0)
		{
			QMessageBox msgBox;

			msgBox.setWindowTitle("Error");
			msgBox.setText("Please Set User Name and Password before enabling Web Management interface");
			msgBox.move(mainpos.x() + 50, mainpos.y() + 300);
			msgBox.exec();

			Button->setChecked(false);
			return;
		}

		if (TelnetPort == 0)
		{
			createTelnetConfig(nextPortnum++);
			refreshPorts();
		}
		Button->setEnabled(false);
		return;
	}

	if (strcmp(Name, "Update") == 0)
	{
		CreateConfig();
		return;
	}


	int p = 0;

	for (int i = 0; i < 32; i++)
	{
		if (Button == gPortEdit[i])
		{
			editPort(p);
			return;
		}

		p++;

		while (portConfig[p] == nullptr)
			p++;
	}

	for (int i = 0; i < 32; i++)
	{
		if (Button == gapplEdit[i])
		{
			saveAppl(i);
			return;
		}

		p++;

		while (portConfig[p] == nullptr)
			p++;
	}

	QMessageBox msgBox;
	msgBox.setWindowTitle("MessageBox Title");
	msgBox.setText("You Clicked " + ((QPushButton*)sender())->objectName());
	msgBox.move(mainpos.x() + 200, mainpos.y() + 300);
	msgBox.exec();
}
	
void BPQConfigGen::saveGenerations(char * FN)
{
	// Save 6 backups
	
	char FN1[256];
	QFile File(FN);

	if (File.exists())
	{
		QDateTime DateTime = QDateTime::currentDateTime();
		QString Time = DateTime.toString("yyyyMMddHHmmss");
		QByteArray ba = Time.toLatin1();
		sprintf(FN1, "%s.%s", FN, ba.data());
		File.rename(FN, FN1);
	}
}


void BPQConfigGen::ReadConfigFile()
{
	char FN[260];
	FILE * fp;
	FILE * savefp = nullptr;

	char line[512];
	char linecopy[512];

	char * Context;
	bool Comment = false;
	bool inPort = false;		// parsing a PORT section
	bool inBlock = false;		// parsing a section ending in ***
	bool gotUser = false;

	char tempPort[32768] = "";

	mainConfig[0] = 0;

	nextPortnum = 0;
	activePortnum = 0;
	nextApplNum = 0;
	NextUser = 0;

	for (int i = 0; i < 32; i++)
	{
		portSetVisible(i, 0);

		if (portConfig[i])
		{
			free(portConfig[i]);
			portConfig[i] = nullptr;
		}

		Appls[i].Num = 0;
		Appls[i].Command[0] = 0;
	}

	ui.BBS->setEnabled(true);
	ui.CHAT->setEnabled(true);
	ui.RMS->setEnabled(true);
	ui.WEBMGMT->setEnabled(true);
	ui.BBS->setChecked(false);
	ui.CHAT->setChecked(false);
	ui.RMS->setChecked(false);
	ui.WEBMGMT->setChecked(false);


	sprintf(FN, "%s/bpq32.cfg", ui.Directory->text().toLocal8Bit().constData());

	if ((fp = fopen(FN, "r")) == NULL)
	{
		QMessageBox msgBox;
		msgBox.setWindowTitle("Error");
		msgBox.setText("Failed to open bpq32.cfg");
		msgBox.move(mainpos.x() + 200, mainpos.y() + 300);
		msgBox.exec();
		return;
	}

	// Read first line and check that writted by us

	if (fgets(line, 512, fp) == NULL)
	{
		QMessageBox msgBox;
		msgBox.setWindowTitle("Error");
		msgBox.setText("Empty bpq32.cfg");
		msgBox.move(mainpos.x() + 200, mainpos.y() + 300);
		msgBox.exec();
		return;
	}

	if (strstr(line, "; Written by BPQConfigGen") == 0)
	{
		// Warn
		int ret = QMessageBox::warning(this, tr("Warning!!"),
			tr("File not written by BPQConfigGen. You can continue but file may not"
				" be read correctly and may be corrupted if saved"),
			QMessageBox::Ok | QMessageBox::Cancel, QMessageBox::Cancel);

		if (ret == QMessageBox::Cancel)
			return;
	}

	HadConfig = true;		// we've read a config 

	while (1)
	{
		char * ret;

		ret = fgets(line, 512, fp);
		strcpy(linecopy, line);			// Save original

		if (ret == NULL)
		{
			if (savefp)
			{
				// we have reached eof on an include file - switch back

				fclose(fp);
				fp = savefp;
				savefp = NULL;
				continue;
			}

			break;		// end of config
		}

		char * ptr;

		ptr = strtok_s(line, " =\t\r\n", &Context);

		// Try keeping comments



		if (ptr == NULL || ptr[0] == 0 || ptr[0] == ';')
		{
			// Comment

			if (inPort)
				strcat(tempPort, linecopy);
			else
				strcat(mainConfig, linecopy);

			continue;
		}

		if (strlen(ptr) > 1)
		{
			if (memcmp(ptr, "/*", 2) == 0)
			{
				Comment = TRUE;
				strcat(mainConfig, linecopy);
				continue;
			}
			else if (memcmp(ptr, "*/", 2) == 0)
			{
				Comment = FALSE;
				strcat(mainConfig, linecopy);
				continue;
			}
		}

		if (Comment)
		{
			strcat(mainConfig, linecopy);
			continue;
		}

		strlop(Context, ';');

		if (stricmp(ptr, "#include") == 0)
		{
			savefp = fp;

			strlop(Context, 10);
			strlop(Context, 13);

			sprintf(FN, "%s/%s", ui.Directory->text().toLocal8Bit().constData(), Context);

			if ((fp = fopen(FN, "r")) == NULL)
			{
				QMessageBox msgBox;
				msgBox.setWindowTitle("Failed to open include file");
				msgBox.setText(FN);
				msgBox.move(mainpos.x() + 200, mainpos.y() + 300);
				msgBox.exec();

				fp = savefp;
				savefp = NULL;
			}

			continue;			// get next line
		}

		if (inBlock)
		{
			strcat(mainConfig, linecopy);

			if (memcmp(ptr, "***", 3) == 0)
				inBlock = FALSE;

			continue;
		}

		if (inPort)
		{
			// Dont add USER records to port config so don't add to tempPort

			if (stricmp(ptr, "USER") == 0)
			{
				if (PortType[activePortnum] == H_TELNET)
				{
					struct UserRec * USER = &Users[NextUser];
					char *User, *Pwd, *UserCall, *Secure, *Appl;
					char Param[8][256];
					char * ptr1, *ptr2;
					int n = 0;

					// can't use strtok as may have ,,

					// USER=user,password,call,appl,SYSOP

					memset(Param, 0, 2048);
					strlop(Context, 13);
					strlop(Context, ';');

					ptr1 = Context;

					while (ptr1 && *ptr1 && n < 8)
					{
						ptr2 = strchr(ptr1, ',');
						if (ptr2) *ptr2++ = 0;

						strcpy(&Param[n][0], ptr1);
						strlop(Param[n++], ' ');
						ptr1 = ptr2;
						while (ptr1 && *ptr1 && *ptr1 == ' ')
							ptr1++;
					}

					User = &Param[0][0];

					Pwd = &Param[1][0];
					UserCall = &Param[2][0];
					Appl = &Param[3][0];
					Secure = &Param[4][0];

					if (User[0] == 0 || Pwd[0] == 0 || UserCall[0] == 0) // invalid record
						continue;

					USER->Callsign = strdup(UserCall);
					USER->Password = strdup(Pwd);
					USER->UserName = strdup(User);
					USER->Appl = (char *)malloc(32);
					USER->Appl[0] = 0;
					USER->Secure = FALSE;

					if (stricmp(Secure, "SYSOP") == 0)
						USER->Secure = TRUE;

					if (Appl[0] && strcmp(Appl, "\"\"") != 0)
					{
						strcpy(USER->Appl, Appl);
						strcat(USER->Appl, "\r\n");
					}

					NextUser++;

					if (gotUser == FALSE)	// only on first USER line
					{
						ui.sysopUser->setText(User);
						ui.sysopPass->setText(Pwd);

						ui.WEBMGMT->setChecked(true);
						ui.WEBMGMT->setEnabled(false);

						gotUser = TRUE;
					}
				}
				continue;
			}

			strcat(tempPort, linecopy);

			if (stricmp(ptr, "ENDPORT") == 0)
			{
				portConfig[activePortnum] = strdup(tempPort);
				inPort = FALSE;
				portSetVisible(activePortnum, 1);
				gPortType[activePortnum]->setText(Types[PortType[activePortnum]]);

//				if (Param1[activePortnum][0])
//					gParam1[activePortnum]->setText(Param1[activePortnum]);

//				if (Param2[activePortnum])
//					gParam2[activePortnum]->setText(QString::number(Param2[activePortnum]));

				if (activePortnum == nextPortnum)  // PORTNUM Not Used
					nextPortnum++;

				while (portConfig[nextPortnum] && nextPortnum < 31)
					nextPortnum++;

				continue;
			}

			if (stricmp(ptr, "PORTNUM") == 0)
			{
				char * num = strtok_s(NULL, " =\t\r\n", &Context);
				int i = atoi(num);

				if (i > 0 && i < 32)
					activePortnum = i - 1;

				continue;
			}
				
			if (stricmp(ptr, "ID") == 0)
			{
				char * id = strtok_s(NULL, "\r\n;", &Context);

				if (strlen(id) > 31)
					id[31] = 0;

				strcpy(ID[activePortnum], id);

				gPortName[activePortnum]->setText(id);
				continue;
			}

			if (stricmp(ptr, "TYPE") == 0)
			{
				char * type = strtok_s(NULL, " =\t\r\n", &Context);

				if (stricmp(type, "ASYNC") == 0)
				{
					PortType[activePortnum] = H_KISS;
				}

				if (stricmp(type, "LOOPBACK") == 0 || stricmp(type, "INTERNAL") == 0)
				{
					PortType[activePortnum] = H_LOOP;
				}

				continue;

			}

			if (stricmp(ptr, "IPADDR") == 0 || stricmp(ptr, "COMPORT") == 0 || stricmp(ptr, "IOADDR") == 0)
			{
				char * addr = strtok_s(NULL, " =\t\r\n", &Context);

				strcpy(Param1[activePortnum], addr);
				continue;
			}

			if (stricmp(ptr, "SPEED") == 0)
			{
				char * addr = strtok_s(NULL, " =\t\r\n", &Context);

				Param2[activePortnum] = atoi(addr);
				continue;
			}

			if (stricmp(ptr, "TCPPORT") == 0)
			{
				char * addr = strtok_s(NULL, " =\t\r\n", &Context);

				Param2[activePortnum] = atoi(addr);

				if (PortType[activePortnum] == H_KISS)
					PortType[activePortnum] = H_KISSTCP;

				continue;
			}

			if (stricmp(ptr, "UDPPORT") == 0)
			{
				char * addr = strtok_s(NULL, " =\t\r\n", &Context);

				Param2[activePortnum] = atoi(addr);

				if (PortType[activePortnum] == H_KISS)
					PortType[activePortnum] = H_KISSUDP;

				continue;
			}

			if (stricmp(ptr, "ADDR") == 0)
			{
				// ADDR line - Host then Port

				char * addr = strtok_s(NULL, " =\t\r\n", &Context);
				strcpy(Param1[activePortnum], addr);

				addr = strtok_s(NULL, " =\t\r\n", &Context);
				Param2[activePortnum] = atoi(addr);

				continue;
			}


			if (stricmp(ptr, "CMSCALL") == 0)
			{
				ui.CMSCALL->setText(strtok_s(NULL, " =\t\r\n", &Context));
				continue;
			}

			if (stricmp(ptr, "CMSPASS") == 0)
			{
				ui.CMSPASS->setText(strtok_s(NULL, " =\t\r\n", &Context));
				continue;
			}

			if (stricmp(ptr, "DRIVER") == 0 || stricmp(ptr, "DLLNAME") == 0)
			{
				ptr = strtok_s(NULL, ".\t\r\n", &Context);

				if (ptr == NULL)
					continue;

				if (stricmp(ptr, "ARDOP") == 0)
				{
					PortType[activePortnum] = H_ARDOP;
					continue;
				}

				if (stricmp(ptr, "BPQAXIP") == 0)
				{
					PortType[activePortnum] = H_AXIP;
					continue;
				}

				if (stricmp(ptr, "BPQETHER") == 0)
				{
					PortType[activePortnum] = H_ETHER;
					continue;
				}

				if (stricmp(ptr, "BPQVKISS") == 0)
				{
					PortType[activePortnum] = H_VKISS;
					continue;
				}

				if (stricmp(ptr, "FLDIGI") == 0)
				{
					PortType[activePortnum] = H_FLDIGI;
					continue;
				}

				if (stricmp(ptr, "KAMPACTOR") == 0)
				{
					PortType[activePortnum] = H_KAM;
					continue;
				}

				if (stricmp(ptr, "SCSPACTOR") == 0)
				{
					PortType[activePortnum] = H_SCS;
					continue;
				}

				if (stricmp(ptr, "SERIAL") == 0)
				{
					PortType[activePortnum] = H_SERIAL;
					continue;
				}

				if (stricmp(ptr, "TELNET") == 0)
				{
					PortType[activePortnum] = H_TELNET;
					TelnetPort = activePortnum + 1;
					continue;
				}
				if (stricmp(ptr, "UZ7HO") == 0)
				{
					PortType[activePortnum] = H_UZ7HO;
					continue;
				}

				if (stricmp(ptr, "WINMOR") == 0)
				{
					PortType[activePortnum] = H_WINMOR;
					continue;
				}

				if (stricmp(ptr, "VARA") == 0)
				{
					PortType[activePortnum] = H_VARA;
					continue;
				}
			}
			continue;
		}

		if (stricmp(ptr, "IPGATEWAY") == 0 ||
			stricmp(ptr, "PORTMAPPER") == 0 ||
			stricmp(ptr, "APRSDIGI") == 0 ||
			stricmp(ptr, "ROUTES:") == 0 ||
			stricmp(ptr, "INFOMSG:") == 0 ||
			stricmp(ptr, "IDMSG:") == 0 ||
			stricmp(ptr, "BTEXT:") == 0 ||
			stricmp(ptr, "CTEXT:") == 0)

		{
			strcat(mainConfig, linecopy);
			inBlock = TRUE;
			continue;
		}


		if (stricmp(ptr, "NODECALL") == 0)
		{
			NODECALL = strtok_s(NULL, " =\t\r\n", &Context);
			ui.NODECALL->setText(NODECALL);
			continue;
		}

		if (stricmp(ptr, "NODEALIAS") == 0)
		{
			NODEALIAS = strtok_s(NULL, " =\t\r\n", &Context);
			ui.NODEALIAS->setText(NODEALIAS);
			continue;
		}

		if (stricmp(ptr, "LOCATOR") == 0)
		{
			char * ptr = strtok_s(NULL, "\t\r\n", &Context);
			strlop(ptr, ' ');
			LOC = ptr;
			ui.LOC->setText(LOC);
			continue;
		}

		if (stricmp(ptr, "PORT") == 0)
		{
			strcpy(tempPort, linecopy);
			inPort = TRUE;
			activePortnum = nextPortnum;
			continue;
		}

		if (stricmp(ptr, "APPLICATION") == 0)
		{
			struct APPL * pAppl;

			int Num = atoi(strtok_s(NULL, " ,\t\r\n", &Context));
			char * Appl = strtok_s(NULL, " ,\t\r\n", &Context);
			char * CmdAlias = NULL;
			char * ApplCall = NULL;
			char * ApplAlias = NULL;
			char * ApplQual = NULL;

			if (Context[0] == ',')
				Context++;
			else
				CmdAlias = strtok_s(NULL, ",\t\r\n", &Context);

			if (Context[0] == ',')
				Context++;
			else
				ApplCall = strtok_s(NULL, " ,\t\r\n", &Context);

			if (Context[0] == ',')
				Context++;
			else
				ApplAlias = strtok_s(NULL, " ,\t\r\n", &Context);

			ApplQual = strtok_s(NULL, " ,\t\r\n", &Context);

			if (Appl)
			{
				pAppl = &Appls[nextApplNum++];

				pAppl->Num = Num;
				strcpy(pAppl->Command, Appl);
				if (ApplCall)
					strcpy(pAppl->ApplCall, ApplCall);
				if (ApplAlias)
					strcpy(pAppl->ApplAlias, ApplAlias);
				
				if (ApplQual)
					pAppl->ApplQual = atoi(ApplQual);

				if (CmdAlias)
					strcpy(pAppl->CommandAlias, CmdAlias);

				if (strcmp(Appl, "BBS") == 0)
				{
					BBSAppl = Num;
					ui.BBS->setChecked(true);
					ui.BBS->setEnabled(false);

				}

				else if (strcmp(Appl, "RMS") == 0)
				{
					RMSAppl = Num;
					ui.RMS->setChecked(true);
					ui.RMS->setEnabled(false);
				}

				else if (stricmp(Appl, "CHAT") == 0)
				{
					ChatAppl = Num;
					ui.CHAT->setChecked(true);
					ui.CHAT->setEnabled(false);
				}
			}
			continue;
		}

		if (stricmp(line, "SIMPLE") == 0)
			continue;

		if (stricmp(line, "LINMAIL") == 0)
			continue;

		if (stricmp(line, "LINCHAT") == 0)
			continue;

		// Something we don't handle - write to main config section

//		qDebug() << linecopy;
		strcat(mainConfig, linecopy);
	}

	fclose(fp);

	refreshPorts();
	refreshAppls();
}

void BPQConfigGen::WriteARDOPPort(FILE * fp)
{
	QString str3 = ui.LOC->text().toUpper();

	fprintf(fp, "\r\nPORT\r\n");
	fprintf(fp, " ID=ARDOP\r\n");
	fprintf(fp, " DRIVER=ARDOP\r\n");
	fprintf(fp, " INTERLOCK=1\r\n");
	fprintf(fp, " CONFIG\r\n");
	fprintf(fp, "  ADDR 127.0.0.1 8515\r\n");
	fprintf(fp, "  DEBUGLOG TRUE\r\n");
	fprintf(fp, "  CMDTRACE TRUE\r\n");
	if (str3.size())
		fprintf(fp, "  GRIDSQUARE %s\r\n", str3.toLocal8Bit().constData());
	fprintf(fp, "ENDPORT\r\n");
}


void BPQConfigGen::WriteUZ7HOPort(FILE * fp)
{
	fprintf(fp, "\r\nPORT\r\n");
	fprintf(fp, " ID=UZ7HO\r\n");
	fprintf(fp, " DRIVER=UZ7HO\r\n");
	fprintf(fp, " CHANNEL=A\r\n");
	fprintf(fp, " INTERLOCK=1\r\n");
	fprintf(fp, " CONFIG\r\n");
	fprintf(fp, "  ADDR 127.0.0.1 8000\r\n");
	fprintf(fp, "  UPDATEMAP\r\n");
	fprintf(fp, "ENDPORT\r\n");
}

void BPQConfigGen::WriteWINMORPort(FILE *fp)
{
	fprintf(fp, "\r\nPORT\r\n");
	fprintf(fp, " ID=WINMOR\r\n");
	fprintf(fp, " DRIVER=WINMOR\r\n");
	fprintf(fp, " INTERLOCK=1\r\n");
	fprintf(fp, " CONFIG\r\n");
	fprintf(fp, "  ADDR 127.0.0.1 8500\r\n");
	fprintf(fp, "  DEBUGLOG TRUE\r\n");
	fprintf(fp, "  BUSYLOCK TRUE\r\n");
	fprintf(fp, "  BW 1600\r\n");
	fprintf(fp, "ENDPORT\r\n");
}

void BPQConfigGen::WriteVARAPort(FILE * fp)
{
	fprintf(fp, "\r\nPORT\r\n");
	fprintf(fp, " ID=VARA\r\n");
	fprintf(fp, " DRIVER=VARA\r\n");
	fprintf(fp, " INTERLOCK=1\r\n");
	fprintf(fp, " CONFIG\r\n");
	fprintf(fp, "  ADDR 127.0.0.1 8300\r\n");
	fprintf(fp, "ENDPORT\r\n");
}

void BPQConfigGen::AddTCPKISSPort(FILE * fp)
{
	fprintf(fp, "PORT\r\n");

	fprintf(fp, "ID=KISS over TCP Slave\r\n");
	fprintf(fp, "TYPE=ASYNC\r\n");
	fprintf(fp, "IPADDR=0.0.0.0\r\n");
	fprintf(fp, "TCPPORT=8100\r\n");
	fprintf(fp, "ENDPORT\r\n");
	fprintf(fp, "\r\n");
}

void refreshPorts()
{
	int i = 0;

	for (int i = 0; i < 32; i++)
		portSetVisible(i, 0);

	for (int pnum = 0; pnum < 32; pnum++)
	{
		if (portConfig[pnum])
		{
			gPortNum[i]->setText(QString::number(pnum + 1));
			gPortType[i]->setText(Types[PortType[pnum]]);
			gPortName[i]->setText(ID[pnum]);
		
			portSetVisible(i, 1);
			i++;
		}
	}

}

void portSetVisible(int i, int v)
{
	gPortNum[i]->setVisible(v);
	gPortType[i]->setVisible(v);
	gPortEdit[i]->setVisible(v);
//	gParam2[i]->setVisible(v);
	gPortName[i]->setVisible(v);
}

void applSetVisible(int i, int v)
{
	gapplNum[i]->setVisible(v);
	gapplCmd[i]->setVisible(v);
	gapplCmdAlias[i]->setVisible(v);
	gapplCall[i]->setVisible(v);
	gapplAlias[i]->setVisible(v);
	gapplQual[i]->setVisible(v);
	gapplEdit[i]->setVisible(v);
}


void refreshAppls()
{
	int i = 0;

	for (i = 0; i < 32; i++)
		applSetVisible(i, 0);

	for (i = 0; i < 32; i++)
	{
		struct APPL * pAppl = &Appls[i];

		if (pAppl->Command[0])
		{
			gapplNum[i]->setText(QString::number(pAppl->Num));
			gapplCmd[i]->setText(pAppl->Command);
			gapplCmdAlias[i]->setText(pAppl->CommandAlias);
			gapplCall[i]->setText(pAppl->ApplCall);
			gapplAlias[i]->setText(pAppl->ApplAlias);
			gapplQual[i]->setText(QString::number(pAppl->ApplQual));
			applSetVisible(i, 1);
		}
	}
}

void BPQConfigGen::GetParams()
{
	NodeCall = ui.NODECALL->text().toUpper();
	NodeAlias = ui.NODEALIAS->text().toUpper();
	LOC = ui.LOC->text().toUpper();

	User = ui.sysopUser->text();
	Pass = ui.sysopPass->text();

	CMSCall = ui.CMSCALL->text().toUpper();;
	CMSPass = ui.CMSPASS->text();

	RMS = ui.RMS->isChecked();

	strcpy(LoppedCall, NodeCall.toLocal8Bit().constData());

	char * ptr = strchr(LoppedCall, '-');
	if (ptr)
		*ptr = 0;

	if (User.length() || Pass.length())
	{
		// Create or update first User Record

		struct UserRec * USER = &Users[0];

		if (USER->UserName)
			free(USER->UserName);

		if (USER->Password)
			free(USER->Password);

		USER->UserName = strdup(User.toLocal8Bit());
		USER->Password = strdup(Pass.toLocal8Bit());

		if (NextUser == 0)
		{
			// Creating record

			USER->Callsign = strdup(LoppedCall);
			USER->Appl = strdup("");
			USER->Secure = 1;
			NextUser++;
		}
	}
}


void BPQConfigGen::CreateConfig()
{
	QMessageBox msgBox;

	QString str = ui.Directory->text();

	char FN[260];
	FILE *fp;

	GetParams();

	sprintf(FN, "%s/bpq32.cfg", str.toLocal8Bit().constData());

	saveGenerations(FN);

	fp = fopen(FN, "wb");

	if (fp == NULL)
	{
		msgBox.setWindowTitle("Error");
		msgBox.setText("Failed to open bpq32.cfg");
		msgBox.move(mainpos.x() + 200, mainpos.y() + 300);
		msgBox.exec();
		return;
	}

	fprintf(fp, "; Written by BPQConfigGen\r\n");
	fprintf(fp, "\r\n");
	fprintf(fp, "SIMPLE\r\n");
	fprintf(fp, "NODECALL=%s\r\n", NodeCall.toLocal8Bit().constData());

	if (NodeAlias.size())
		fprintf(fp, "NODEALIAS=%s\r\n", NodeAlias.toLocal8Bit().constData());

	if (LOC.size())
		fprintf(fp, "LOCATOR=%s\r\n", LOC.toLocal8Bit().constData());

	fprintf(fp, "\r\n");

	fprintf(fp, mainConfig);

	// Write PORT Lines 

	if (portConfig[0] == nullptr)		// No ports
	{
		QMessageBox msgBox;

		msgBox.setWindowTitle("Warning");
		msgBox.setText("You have created a config without any ports. You will need to add at least one before node will start");
		msgBox.move(mainpos.x() + 50, mainpos.y() + 300);
		msgBox.exec();
	}

	// write ports and applications

	for (int i = 0; i < 32; i++)
	{
		if (portConfig[i])
		{
			if (PortType[i] == H_TELNET)
			{
				// add in USER lines

				char * ptr = strstr(portConfig[i], "ENDPORT");

				if (ptr)
					*ptr = 0;

				fprintf(fp, portConfig[i]);

				for (int i = 0; i < NextUser; i++)
				{
					struct UserRec * USER = &Users[i];
					fprintf(fp, "  USER=%s,%s,%s,%s,%s;\r\n",
						USER->UserName, USER->Password, USER->Callsign, USER->Appl,
						(USER->Secure) ? "SYSOP" : "");
				}

				fprintf(fp, "ENDPORT\r\n");
				strcat(portConfig[i], "ENDPORT\r\n");
			}
			else
			{
				fprintf(fp, portConfig[i]);
				fprintf(fp, "\r\n");
			}
		}
	}

	fprintf(fp, "\r\n");

	for (int i = 0; i < 32; i++)
	{
		struct APPL * pAppl = &Appls[i];

		if (pAppl->Command[0])
			fprintf(fp, "APPLICATION %d,%s,%s,%s,%s,%d\r\n",
				pAppl->Num,
				pAppl->Command,
				pAppl->CommandAlias,
				pAppl->ApplCall,
				pAppl->ApplAlias,
				pAppl->ApplQual);
	}

	fprintf(fp, "\r\n");

	//#ifndef Q_OS_WIN32

	if (ui.BBS->isChecked())
		fprintf(fp, "LINMAIL\r\n");

	if (ui.CHAT->isChecked())
		fprintf(fp, "LINCHAT\r\n");

	//#endif

	fclose(fp);

	refreshPorts();

	msgBox.move(mainpos.x() + 200, mainpos.y() + 300);

	msgBox.setWindowTitle("Save");

	if (HadConfig)
		msgBox.setText("bpq32.cfg updated");
	else
		msgBox.setText("bpq32.cfg created");

	msgBox.exec();
}



QComboBox * pPortType;
QLineEdit * pportNum;
QLineEdit * pParam0;
QLineEdit * pParam1;
QLineEdit * pParam2;
QLineEdit * pParam3;

QLabel * lportNum;
QLabel * pLabel0;
QLabel * pLabel1;
QLabel * pLabel2;
QLabel * pLabel3;

void PortDialog::PortTypeChanged(int Selected)
{
	QVariant Q = pPortType->currentText();

	pParam0->setText(Types[Selected]);

	pLabel1->setVisible(0);
	pParam1->setVisible(0);
	pLabel2->setVisible(0);
	pParam2->setVisible(0);
	pLabel3->setVisible(0);
	pParam3->setVisible(0);


	switch (Selected)
	{
	case H_ARDOP:
	case H_FLDIGI:
	case H_KISSTCP:
	case H_MPSK:
	case H_UZ7HO:
	case H_V4:
	case H_VARA:
	case H_WINMOR:

		pLabel1->setText("Host Name");
		pLabel2->setText("TCP Port");

		pLabel1->setVisible(1);
		pParam1->setVisible(1);
		pLabel2->setVisible(1);
		pParam2->setVisible(1);

		pParam1->setText("127.0.0.1");
		pParam2->setText(QString::number(defaultPort[Selected]));

		break;

	case H_KISSUDP:
		
		pLabel1->setText("Host Name");
		pLabel2->setText("UDP Ports");

		pLabel1->setVisible(1);
		pParam1->setVisible(1);
		pLabel2->setVisible(1);
		pParam2->setVisible(1);

		break;

	case H_KISS:
	case H_AEA:
	case H_HAL:
	case H_KAM:
	case H_SCS:
	case H_TRK:
	case H_TRKM:
	case H_SERIAL:
	case H_VKISS:

		pLabel1->setText("Serial Port");
		pLabel2->setText("Speed");
		pParam2->setText(QString::number(defaultPort[Selected]));
		pLabel1->setVisible(1);
		pParam1->setVisible(1);
		pLabel2->setVisible(1);
		pParam2->setVisible(1);

		break;

	case H_LOOP:

		pParam0->setText("Loopback Port");
		break;

	}

	switch (Selected)
	{
	case H_KISS:
	case H_KISSTCP:
	case H_KISSUDP:
	case H_UZ7HO:

		pLabel3->setVisible(1);
		pParam3->setVisible(1);
		pParam3->setText("A");
	}
}

PortDialog::PortDialog(QWidget *parent) : QDialog(parent)
{
	this->resize(562, 491);

	this->move(mainpos.x() + 50, mainpos.y() + 50);

	setWindowTitle(tr("Add Port"));

	while (portConfig[nextPortnum] && nextPortnum < 31)
		nextPortnum++;

	lportNum = new QLabel("Portnum", this);
	lportNum->setGeometry(QRect(20, 22, 60, 18));
	pportNum = new QLineEdit(this);
	pportNum->setGeometry(QRect(85, 22, 20, 18));
	pportNum->setText(QString::number(nextPortnum + 1));

	pPortType = new QComboBox(this);
	pPortType->setGeometry(QRect(125, 22, 120, 18));

	pLabel0 = new QLabel("ID", this);
	pLabel0->setGeometry(QRect(20, 52, 100, 18));
	pParam0 = new QLineEdit(this);
	pParam0->setGeometry(QRect(125, 52, 300, 18));

	pLabel1 = new QLabel("Serial Port", this);
	pLabel1->setGeometry(QRect(20, 82, 100, 18));
	pParam1 = new QLineEdit(this);
	pParam1->setGeometry(QRect(125, 82, 100, 18));

	pLabel2 = new QLabel("Speed", this);
	pLabel2->setGeometry(QRect(20, 112, 100, 18));
	pParam2 = new QLineEdit(this);
	pParam2->setGeometry(QRect(125, 112, 100, 18));

	pLabel3 = new QLabel("Channel", this);
	pLabel3->setGeometry(QRect(20, 142, 100, 18));
	pParam3 = new QLineEdit(this);
	pParam3->setGeometry(QRect(125, 142, 15, 18));

	pLabel3->setVisible(0);
	pParam3->setVisible(0);

	for (int j = 0; j < NUMBEROFTYPES; j++)
		pPortType->addItem(&Types[j][0]);

	buttonBox = new QDialogButtonBox(QDialogButtonBox::Save | QDialogButtonBox::Cancel, this);
	buttonBox->setGeometry(QRect(400, 400, 100, 18));

	PortTypeChanged(0);			// Set default

	connect(pPortType, SIGNAL(currentIndexChanged(int)), this, SLOT(PortTypeChanged(int)));

	connect(buttonBox, SIGNAL(accepted()), this, SLOT(myaccept()));
	connect(buttonBox, SIGNAL(rejected()), this, SLOT(myreject()));
}


PortDialog::~PortDialog()
{
}

void PortDialog::myaccept()
{
	QVariant Q;
	QString val = pportNum->text();
	QByteArray ba;
	char * ptr;

	int port = atoi(val.toLatin1()) - 1;

	PortType[port] = pPortType->currentIndex();

	val = pParam0->text();
	ba = val.toLatin1();

	strcpy(ID[port], ba.data());

	val = pParam1->text();
	ba = val.toLatin1();
	strcpy(Param1[port], ba.data());

	val = pParam2->text();
	ba = val.toLatin1();
	Param2[port] = atoi(ba.data());

	val = pParam3->text();
	ba = val.toLatin1();
	strcpy(Param3[port], ba.data());

//	gPortType[port]->setText(Types[PortType[port]]);

//	gPortName[port]->setText(ID[port]);

//	if (Param1[port][0])
//		gParam1[port]->setText(Param1[port]);

//	if (Param2[port])
//		gParam2[port]->setText(QString::number(Param2[port]));

	if (portConfig[port])
		free(portConfig[port]);

	ptr = portConfig[port] = (char *)malloc(1000);

	QMessageBox msgBox;

	switch (PortType[port])
	{
	case H_AEA:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=AEAPACTOR\r\n");
		ptr += sprintf(ptr, " COMPORT=%s\r\n", Param1[port]);
		ptr += sprintf(ptr, " SPEED=%d\r\n", Param2[port]);
		ptr += sprintf(ptr, " INTERLOCK=1\r\n");
		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	case H_ARDOP:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=ARDOP\r\n");
		ptr += sprintf(ptr, " INTERLOCK=1\r\n");
		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "  ADDR %s %d\r\n", Param1[port], Param2[port]);
		ptr += sprintf(ptr, "  DEBUGLOG TRUE\r\n");
		ptr += sprintf(ptr, "  CMDTRACE TRUE\r\n");
		ptr += sprintf(ptr, "  CWID TRUE\r\n");
		ptr += sprintf(ptr, "  MAXCONREQ 10\r\n");
		if (LOC.size())
			ptr += sprintf(ptr, "  GRIDSQUARE %s\r\n", LOC.toLocal8Bit().constData());

		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	case H_AXIP:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=BPQAXIP\r\n");
		ptr += sprintf(ptr, " PACLEN=236\r\n");
		ptr += sprintf(ptr, " FRACK=5000\r\n");
		ptr += sprintf(ptr, " RESPTIME=1500\r\n");
		ptr += sprintf(ptr, " MAXFRAME=4\r\n");
		ptr += sprintf(ptr, " RETRIES=6\r\n");
		ptr += sprintf(ptr, " QUALITY=192\r\n");
		ptr += sprintf(ptr, " MINQUAL=191\r\n");

		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "  BROADCAST NODES\r\n");
		ptr += sprintf(ptr, "  MHEARD ON\r\n");
		ptr += sprintf(ptr, "  UDP 10093\r\n");
		ptr += sprintf(ptr, "  MAP XXXXX xxxx.com UDP 10093 B\r\n");
		ptr += sprintf(ptr, "ENDPORT\r\n");

		msgBox.setWindowTitle("AXIP Port Config");
		msgBox.setText("Edit config and replace the sample MAP line with your MAP info");
		msgBox.move(mainpos.x() + 100, mainpos.y() + 300);
		msgBox.exec();

		break;


//	case H_FLDIGI:
//		break;


	case H_HAL:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=HALDRIVER\r\n");
		ptr += sprintf(ptr, " COMPORT=%s\r\n", Param1[port]);
		ptr += sprintf(ptr, " SPEED=%d\r\n", Param2[port]);
		ptr += sprintf(ptr, " INTERLOCK=1\r\n");
		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "  TONES 160/1800\r\n");
		ptr += sprintf(ptr, "  NEEDXONXOFF; Needed if DSP4200\r\n");
		ptr += sprintf(ptr, "  DEFAULTMODE CLOVER\r\n");
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	case H_KAM:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=KAMPACTOR\r\n");
		ptr += sprintf(ptr, " COMPORT=%s\r\n", Param1[port]);
		ptr += sprintf(ptr, " SPEED=%d\r\n", Param2[port]);
		ptr += sprintf(ptr, " INTERLOCK=1\r\n");
		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;



	case H_KISS:
	case H_KISSTCP:
	case H_KISSUDP:
	case H_VKISS:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);

		if ((PortType[port]) == H_VKISS)
		{
			ptr += sprintf(ptr, "DRIVER=BPQVKISS\r\n");
		}
		else
		{
			ptr += sprintf(ptr, " TYPE=ASYNC\r\n");
			ptr += sprintf(ptr, " PROTOCOL=KISS\r\n");
		}

		if ((PortType[port]) == H_KISS || (PortType[port]) == H_VKISS)
		{
			ptr += sprintf(ptr, " COMPORT=%s\r\n", Param1[port]);
			ptr += sprintf(ptr, " SPEED=%d\r\n", Param2[port]);
		}
		else if ((PortType[port]) == H_KISSTCP)
		{
			ptr += sprintf(ptr, " IPADDR=%s\r\n", Param1[port]);
			ptr += sprintf(ptr, " TCPPORT=%d\r\n", Param2[port]);
		}
		else
		{
			ptr += sprintf(ptr, " IPADDR=%s\r\n", Param1[port]);
			ptr += sprintf(ptr, " UDPPORT=%d", Param2[port]);
		}

		ptr += sprintf(ptr, " CHANNEL=%s\r\n", Param3[port]);
		ptr += sprintf(ptr, " PACLEN=128\r\n");
		ptr += sprintf(ptr, " TXDELAY=500\r\n");
		ptr += sprintf(ptr, " FRACK=7000\r\n");
		ptr += sprintf(ptr, " RESPTIME=1500\r\n");
		ptr += sprintf(ptr, " MAXFRAME=4\r\n");
		ptr += sprintf(ptr, " RETRIES=6\r\n");

		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	case H_LOOP:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " TYPE=LOOPBACK\r\n");
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

//	case H_MPSK:

//		break;

	case H_SCS:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=SCSPACTOR\r\n");
		ptr += sprintf(ptr, " COMPORT=%s\r\n", Param1[port]);
		ptr += sprintf(ptr, " SPEED=%d\r\n", Param2[port]);
		ptr += sprintf(ptr, " INTERLOCK=1\r\n");
		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	case H_SERIAL:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=SERIAL\r\n");
		ptr += sprintf(ptr, " COMPORT=%s\r\n", Param1[port]);
		ptr += sprintf(ptr, " SPEED=%d\r\n", Param2[port]);
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	case H_TRK:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=SCSTRACKER\r\n");
		ptr += sprintf(ptr, " COMPORT=%s\r\n", Param1[port]);
		ptr += sprintf(ptr, " SPEED=%d\r\n", Param2[port]);
		ptr += sprintf(ptr, " INTERLOCK=1\r\n");
		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "  USEAPPLCALLS; Accept connects to all APPLCALLS\r\n");
		ptr += sprintf(ptr, "  %%F 1700; Centre Freq for Normal Packet(Default is 1500)\r\n");
		ptr += sprintf(ptr, "  %%B 300;  Radio Baud Rate for Normal Packet\r\n");
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	case H_TRKM:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=TRKMULTI\r\n");
		ptr += sprintf(ptr, " COMPORT=%s\r\n", Param1[port]);
		ptr += sprintf(ptr, " SPEED=%d\r\n", Param2[port]);
		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "  USEAPPLCALLS; Accept connects to all APPLCALLS\r\n");
		ptr += sprintf(ptr, "  %%F 1700; Centre Freq for Normal Packet(Default is 1500)\r\n");
		ptr += sprintf(ptr, "  %%B 1200;  Radio Baud Rate for Normal Packet\r\n");
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	case H_TELNET:

		createTelnetConfig(port);
		break;

	case H_UZ7HO:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=UZ7HO\r\n");
		ptr += sprintf(ptr, " CHANNEL=%s\r\n", Param3[port]);
		ptr += sprintf(ptr, " PACLEN=128\r\n");
		ptr += sprintf(ptr, " INTERLOCK=1\r\n");
		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "  ADDR %s %d\r\n", Param1[port], Param2[port]);
		ptr += sprintf(ptr, "  MAXSESSIONS 5\r\n");
		ptr += sprintf(ptr, "  UPDATEMAP\r\n");
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	case H_V4:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=V4\r\n");
		ptr += sprintf(ptr, " INTERLOCK=1\r\n");
		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "  ADDR %s %d\r\n", Param1[port], Param2[port]);
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	case H_VARA:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=VARA\r\n");
		ptr += sprintf(ptr, " INTERLOCK=1\r\n");
		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "  ADDR %s %d\r\n", Param1[port], Param2[port]);
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	case H_WINMOR:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, " DRIVER=WINMOR\r\n");
		ptr += sprintf(ptr, " INTERLOCK=1\r\n");
		ptr += sprintf(ptr, " CONFIG\r\n");
		ptr += sprintf(ptr, "  ADDR %s %d\r\n", Param1[port], Param2[port]);
		ptr += sprintf(ptr, "  DEBUGLOG TRUE\r\n");
		ptr += sprintf(ptr, "  BUSYLOCK TRUE\r\n");
		ptr += sprintf(ptr, "  CWID TRUE\r\n");
		ptr += sprintf(ptr, "  BW 1600\r\n");
		ptr += sprintf(ptr, "ENDPORT\r\n");

		break;

	default:

		ptr += sprintf(ptr, "PORT\r\n");
		ptr += sprintf(ptr, " PORTNUM=%d\r\n", port + 1);
		ptr += sprintf(ptr, " ID=%s\r\n", ID[port]);
		ptr += sprintf(ptr, "ENDPORT\r\n");

		msgBox.setWindowTitle("Port Config");
		msgBox.setText("Autoconfig not supported - Edit manually");
		msgBox.move(mainpos.x() + 200, mainpos.y() + 300);
		msgBox.exec();
	}

	refreshPorts();

	if (port == nextPortnum)  // PORTNUM Not Used
		nextPortnum++;

	PortDialog::accept();
}

void PortDialog::myreject()
{
	PortDialog::reject();
}

void BPQConfigGen::editPort(int i)
{
	if (portConfig[i])
	{
		editPortDialog dialog(i);
		dialog.exec();
	}
}


QTextEdit * textEdit;
int editPort;

editPortDialog::editPortDialog(int n, QWidget *parent) : QDialog(parent)
{
	this->resize(500, 400);

	this->move(mainpos.x() + 50, mainpos.y() + 50);

	setWindowTitle(tr("Edit Port"));

	textEdit = new QTextEdit(this);
	textEdit->setText(portConfig[n]);

	textEdit->setGeometry(QRect(5, 5, 490, 370));

	if (PortType[n] == H_TELNET)
	{
		int i;

		this->resize(500, 600);

		UserLabel = new QLabel("   User                   Password                Callsign        Appl", this);
		UserLabel->setGeometry(QRect(5, 380, 490, 20));

		userscrollArea = new QScrollArea(this);
		userscrollArea->setObjectName(QString::fromUtf8("scrollArea"));
		userscrollArea->setGeometry(QRect(5, 405, 490, 170));
		userscrollArea->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
		userscrollArea->setWidgetResizable(false);
		userScroll = new QWidget();
		userScroll->setObjectName(QString::fromUtf8("portScroll"));
		userScroll->setGeometry(QRect(0, 0, 450, 2000));
		userscrollArea->setWidget(userScroll);

		int spares = 0;

		for (i = 0; i < 1024 && spares < 20; i++)
		{
			struct UserRec * USER = &Users[i];

			USER->gUserName = new QLineEdit(userScroll);
			USER->gUserName->setGeometry(QRect(5, 22 * i + 5, 95, 18));

			USER->gPassword = new QLineEdit(userScroll);
			USER->gPassword->setGeometry(QRect(105, 22 * i + 5, 95, 18));

			USER->gCallsign = new QLineEdit(userScroll);
			USER->gCallsign->setGeometry(QRect(205, 22 * i + 5, 75, 18));


			USER->gAppl = new QLineEdit(userScroll);
			USER->gAppl->setGeometry(QRect(285, 22 * i + 5, 95, 18));

			USER->gSYSOP = new QCheckBox("SYSOP", userScroll);
			USER->gSYSOP->setGeometry(QRect(385, 22 * i + 5, 70, 18));

			if (USER->UserName)
			{
				USER->gUserName->setText(USER->UserName);
				USER->gPassword->setText(USER->Password);
				USER->gCallsign->setText(USER->Callsign);
				USER->gAppl->setText(USER->Appl);
				USER->gSYSOP->setChecked(USER->Secure);
			}
			else
				spares++;

		}

		userScroll->setGeometry(QRect(0, 0, 450, i * 22 + 10));

		buttonBox = new QDialogButtonBox(QDialogButtonBox::Save | QDialogButtonBox::Cancel, this);
		buttonBox->setGeometry(QRect(250, 580, 100, 18));
	}

	else
	{
		buttonBox = new QDialogButtonBox(QDialogButtonBox::Save | QDialogButtonBox::Cancel, this);
		buttonBox->setGeometry(QRect(200, 380, 100, 18));
	}

	editPort = n;

	connect(buttonBox, SIGNAL(accepted()), this, SLOT(myaccept()));
	connect(buttonBox, SIGNAL(rejected()), this, SLOT(myreject()));
}




void editPortDialog::myaccept()
{
	QString qs = textEdit->toPlainText();
	QByteArray qb = qs.toLocal8Bit();

	free(portConfig[editPort]);
	portConfig[editPort] = (char *)malloc(qb.length() + 2);
	memcpy(portConfig[editPort], qb.data(), qb.length());
	portConfig[editPort][qb.length()] = 0;

	// if Telnet Port get the User Config

	if (PortType[editPort] == H_TELNET)
	{
		QByteArray ba;
		QString val;
		int i;
		NextUser = 0;

		for (i = 0; i < 1024; i++)
		{
			struct UserRec * USER = &Users[i];

			val = USER->gUserName->text();
			ba = val.toLatin1();

			if (ba.length() == 0)
				break;

			if (USER->UserName)
				free(USER->UserName);

			USER->UserName = strdup(ba.data());

			val = USER->gPassword->text();
			ba = val.toLatin1();

			if (USER->Password)
				free(USER->Password);

			USER->Password = strdup(ba.data());

			val = USER->gCallsign->text();
			ba = val.toLatin1();

			if (USER->Callsign)
				free(USER->Callsign);

			USER->Callsign = strdup(USER->gCallsign->text().toUtf8().toUpper());
			USER->gCallsign->setText(USER->Callsign);

			val = USER->gAppl->text();
	
			if (USER->Appl)
				free(USER->Appl);

			USER->Appl = strdup(ba.data());

			USER->Secure = USER->gSYSOP->isChecked();
		}
		NextUser = i;
	}

	editPortDialog::accept();
}

void editPortDialog::myreject()
{
	editPortDialog::reject();
}

editPortDialog::~editPortDialog()
{
}

void BPQConfigGen::saveAppl(int i)
{
	struct APPL * pAppl = &Appls[i];
	QByteArray ba;

	QString val = gapplNum[i]->text();
	ba = val.toLatin1();
	pAppl->Num = atoi(ba.data());

	val = gapplCmd[i]->text().toUpper();
	ba = val.toLatin1();
	strcpy(pAppl->Command, ba.data());

	val = gapplCmdAlias[i]->text();
	ba = val.toLatin1();
	strcpy(pAppl->CommandAlias, ba.data());

	val = gapplCall[i]->text().toUpper();
	ba = val.toLatin1();
	strcpy(pAppl->ApplCall, ba.data());
	
	val = gapplAlias[i]->text().toUpper();
	ba = val.toLatin1();
	strcpy(pAppl->ApplAlias, ba.data());

	val = gapplQual[i]->text();
	ba = val.toLatin1();
	pAppl->ApplQual = atoi(ba.data()); 

	refreshAppls();
}



editOtherDialog::editOtherDialog(int n, QWidget *parent) : QDialog(parent)
{
	this->resize(610, 510);

	this->move(mainpos.x() + 50, mainpos.y() + 50);

	setWindowTitle(tr("Edit Main Config"));

	textEdit = new QTextEdit(this);
	textEdit->setText(mainConfig);
	textEdit->setGeometry(QRect(5, 5, 600, 470));

	buttonBox = new QDialogButtonBox(QDialogButtonBox::Save | QDialogButtonBox::Cancel, this);
	buttonBox->setGeometry(QRect(250, 485, 100, 18));

	editPort = n;

	connect(buttonBox, SIGNAL(accepted()), this, SLOT(myaccept()));
	connect(buttonBox, SIGNAL(rejected()), this, SLOT(myreject()));
}




void editOtherDialog::myaccept()
{
	QString qs = textEdit->toPlainText();
	QByteArray qb = qs.toLocal8Bit();

	memcpy(mainConfig, qb.data(), qb.length());
	mainConfig[qb.length()] = 0;

	editOtherDialog::accept();
}

void editOtherDialog::myreject()
{
	editOtherDialog::reject();
}

editOtherDialog::~editOtherDialog()
{
}




